c***************************************************************************
c***************************************************************************
c***************************************************************************
c***************************************************************************
c***************************************************************************
c***************************************************************************
c***************************************************************************
c***************************************************************************
c***************************************************************************
        module Global
c****************************************************************************
c       This module should appear in all the main blocks of the code.  
c       It replaces the common block and declares and passes all the 
c       global variables and parameters (those globally accessible).
c       Note that many of the vaiables are allocatable. 
c****************************************************************************
        implicit none
        save
c**************************************************************************** 
c       set some parameters.
c**************************************************************************** 

c**************************************************************************** 
c       character parameters.
c**************************************************************************** 
c        character(1), parameter :: Nstring='N',Cstring='C'

c**************************************************************************** 
c       integer parameters.
c**************************************************************************** 
!     -----------------------------------------------------
!     portable kind values, dp may not be 8 on all machines
!     -----------------------------------------------------
      integer, parameter :: dp = kind(1.0d0)
      integer, parameter :: sp = kind(1.0)
      integer, parameter :: wp = dp

        
c**************************************************************************** 
c       logical parameters.
c**************************************************************************** 
      logical, parameter :: isCS=.false.        ! test the crossing symmetries  
      logical, parameter :: isflex=.false.        ! FLEX-like approximation
      logical, parameter :: is2nd =.false.        ! second order approximation
      logical, parameter :: is_use_new = .false.    ! use FT of g0(tau)
      logical, parameter :: isLEV=.false.        ! calculate the Leading Eigen-Values of chi*g_r  
      logical, parameter :: is_bs_pp=.false.        ! use f_s & f_t to calculate f_d & f_m
                                                 ! try to avoid the instability problem produced in solving
                                                 ! B.S. eq in p-h channel 
      logical, parameter :: use_min_memory = .false.
      logical, parameter :: is_single_line = .true.



c**************************************************************************** 
c       integer parameters.
c**************************************************************************** 
      integer, parameter :: sig_mode = 2       ! use p-h or p-p or both channel vertex (0/1/2) 
                                               ! to calculate self-energy 
      integer, parameter :: sigma0_mode = 2    ! use 0 / 2nd-order /SC 2nd-order / FLEX / exact (0/1/2/3/4) 
                                               ! to initialize the self-energy 


c**************************************************************************** 
c       real parameters.
c**************************************************************************** 
      real(kind=wp), parameter :: pi = 3.14159265358979324_wp
      real(kind=wp), parameter ::  halfr=0.500000000000000_wp
      real(kind=wp), parameter ::   oner=1.000000000000000_wp
      real(kind=wp), parameter ::   twor=2.000000000000000_wp
      real(kind=wp), parameter :: threer=3.000000000000000_wp
      real(kind=wp), parameter ::  fourr=4.000000000000000_wp
c      real(kind=wp), parameter :: n_a_not=1.0    ! we need to stick to the falf-filling case temporarily


c**************************************************************************** 
c       complex parameters.
c**************************************************************************** 
      complex(kind=wp), parameter :: zeroc=(0.0_wp,0.0_wp)
      complex(kind=wp), parameter :: onec=(1.0_wp,0.0_wp)
      complex(kind=wp), parameter :: ii=(0.0_wp,1.0_wp)


      logical :: converged, converged_inner

      integer :: i,j,k,iter,iter_inner,dw, wbar, n, n1, n2, n3, m, m1,   &
     &     kx1,kx2,ky1,ky2,k1,k2,k3,kx,ky,dtype, ztype
      integer :: nkx, nky, nc, nf, nt, nb
      integer :: npt, iter_max, iter_max_inner
      integer :: ierr, rank, nproc
      integer :: k_0, k_pi0, k_pi, p_0, p_pi,p_pi0

      real(kind=wp) :: n_a, mu, dt, du, er,er_t,er_t1,er_t2,             &
     &     dfac_g, dfac_s, alpha0, U, T, eigen_value, led(0:2,0:3),      &
     &     er_min, er_min2, val(1:8)

      complex(kind=wp) :: gm0, gm0_old



c*********************************************
c
c           allocatable arrays
c
c*********************************************

      integer, dimension(:), allocatable ::                               &
     &     inv_f, inv_k, inv, indx, minux, indx1, indx2

      integer, dimension(:,:), allocatable ::                             &
     &     add_ff_f, add_fb_f, dif_f, add_k, dif_k, add_ff, add_fb, dif 


      real(kind=wp), dimension(:), allocatable :: tau
      real(kind=wp), dimension(:,:), allocatable :: epsk, fermi


      complex(kind=wp), dimension(:), allocatable ::                      &
     &     w0, w, sigma, sigma_old, sigma_exact, g_exact, g, g0, gtt,     &
     &     delta, dsigma, delta_sigma    

      complex(kind=wp), dimension(:,:), allocatable ::                    &
     &     chi0ph,chi0pp, fchi0_d, fchi0_m, fchi0_s, fchi0_t 

      complex(kind=wp), dimension(:,:,:), allocatable ::                  &
     &     g_d, g_m, g_s, g_t,f_d, f_m, f_s, f_t,                         &
     &     phi_d, phi_m, psi_s, psi_t, temat, mat1, mat2, mat3, mat4

c**************************************************************************** 
c       Table of Variables
c**************************************************************************** 
c 
c       U:   Coulomb interaction
c       T:   Temperature
c       Nc:  number of sites, Nc = nkx*nky
c       nf:  number of Matsubara frequences
c       nt:  number of moment-frequency points, nt=Nc*nf
c       
c       iter_max: maximum number of iterations for the outer loop
c       iter_max_inner: maximum number of iterations for the inner loop
c       npt: maximum number of iterations for the self-consistent 2nd-order calculation
c       er_min:  convergening criterion of the outer loop (Sigma(0))
c       er_min2: convergening criterion of the inner loop (Gamma(0,0,0))
c       
c       delta: knonecker delta function
c       chi0ph: non-perturbed p-h susceptibility
c       chi0pp: non-perturbed p-p susceptibility
c
c       f_d:  reducible vertex (density channel)
c       f_m:  reducible vertex (magnetic channel)
c       f_s:  reducible vertex (singlet channel)
c       f_t:  reducible vertex (triplet channel)
c
c       g_d:  irreducible vertex (density channel)
c       g_m:  irreducible vertex (magnetic channel)
c       g_s:  irreducible vertex (singlet channel)
c       g_t:  irreducible vertex (triplet channel)
c
c**************************************************************************** 

      end module Global


