      program parquet
c**********************************************************************
c     This is the fortran code of the parquet approximation, in which
c     the fully irreducible vertex Lambda is approximated by the bare
c     vertex v.
c
c
c     Ref: ../notes/: 
c
c**********************************************************************
      use Global
      use mod_tprof
      implicit none
      include 'mpif.h'


      call mpi_init(ierr)
      call mpi_comm_rank(mpi_comm_world,rank,ierr)
      call mpi_comm_size(mpi_comm_world,nproc,ierr)

      call profinit()
      call profstart('pa')

      call readin
      call allocate_array


      ! construct the tables
      ! for frequency
      ! table for mapping the frequencies outside [0,nf) back to it
      do n=0,nf-1
         indx(n-nf) = n
         indx(n) = n
         indx(n+nf) = n
      end do
      ! tables for addtion, subtraction and inversion
      do n1=0,nf-1
         inv_f(n1) = indx(-n1-1)
         do n2=0,nf-1
            add_ff_f(n1,n2) = indx(n1+n2+1)
            add_fb_f(n1,n2) = indx(n1+n2)
            dif_f(n1,n2) = indx(n1-n2)
         end do
      end do

      ! for k-space
      ! table for mapping the momenta outside [0,2*pi) back to it
      ! namely, we use the periodic boundary conditions here
      do i=0,nkx-1
         indx1(i-nkx) = i
         indx1(i) = i
         indx1(i+nkx) = i
      end do
      do i=0,nky-1
         indx2(i-nky) = i
         indx2(i) = i
         indx2(i+nky) = i
      end do
      ! tables for addtion, subtraction and inversion
      do ky1=0,nky-1
         do kx1=0,nkx-1
            k1 = kx1+nkx*ky1
            inv_k(k1) = indx1(-kx1)+nkx*indx2(-ky1)
            do ky2=0,nky-1
               do kx2=0,nkx-1
                  k2 = kx2+nkx*ky2
                  add_k(k1,k2) = indx1(kx1+kx2)+nkx*indx2(ky1+ky2)
                  dif_k(k1,k2) = indx1(kx1-kx2)+nkx*indx2(ky1-ky2)
               end do
            end do
         end do
      end do

      ! for w-k space
      do k1=0,nc-1
         do n1=0,nf-1
            i = n1+nf*k1
            inv(i) = inv_f(n1)+nf*inv_k(k1)
            do k2=0,nc-1
               do n2=0,nf-1
                  j = n2+nf*k2
                  add_ff(i,j) = add_ff_f(n1,n2)+nf*add_k(k1,k2)
                  add_fb(i,j) = add_fb_f(n1,n2)+nf*add_k(k1,k2)
                  dif(i,j) = dif_f(n1,n2)+nf*dif_k(k1,k2)
               end do
            end do
         end do
      end do

      ! kronecker delta function
      delta=0.0
      delta(0) = 1.0



      dt = 1.0_wp/T/real(nf,wp)
      do i=0,nf
         if(is_use_new)then
            tau(i) = (i+0.5)*dt
         else
            tau(i) = i*dt
         end if
      end do

      ! dispersion
      if(nc.eq.1)then           ! single site
         epsk = 0.0
      else if(nky.eq.1)then     ! 1-D chain
         do kx1=0,Nkx-1
            epsk(kx1,0) = -0.5*cos(kx1*2.0*pi/Nkx)
         end do
      else                      ! 2-D
         do kx1=0,Nkx-1
            do ky1=0,Nky-1
               epsk(kx1,ky1) = -0.5*(cos(kx1*2.0*pi/Nkx)
     &                                    +cos(ky1*2.0*pi/Nky))
            end do
         end do
      end if




      do n=0,nf-1
	 if(n.lt.nf/2)then
            w0(n) = (2.0_wp*n+1.0_wp)*pi*T
            w(n) = (2._wp*n+1.0_wp)*pi*T*ii
	 else
            w0(n) = (2.0_wp*(n-nf)+1.0_wp)*pi*T
            w(n) = (2.0_wp*(n-nf)+1.0_wp)*pi*T*ii
	 end if
      end do




      !   exact result for single-site system
      if(rank.eq.0 .and. nc.eq.1)then
	 !  exact result
         do n=0,nf-1
            gtt(n) = -0.5*(exp(tau(n)*U/2.0)+exp(U/2.0/T-tau(n)*U/2.0))
     &                    /(1.0+exp(U/2.0/T))
         end do
         call output('gt_exact',gtt)
         !  using FT with high frequency conditioning
         do n=0,nf-1
            sigma_exact(n) = U**2/(4.0*w(n))
            g_exact(n) = 1.0/(w(n)-U**2/(4.0*w(n)))
         end do
         call cal_gt_hfc(gtt,g_exact)
         call output('gt_exact2',gtt)
      end if

c
c          initialization of one-particle quantities
c

      call cal_g0(g0)


      if(rank.eq.0)then
         call sc2nd(sigma,0.0001_wp,npt)
         g = g0/(1.0-g0*sigma)
         if(is_use_new)then
            call cal_gt(gtt,g0)
            call output('gt0',gtt)
            call cal_gt(gtt,g)
            call output('gt2',gtt)
         else
            call cal_gt_hfc(gtt,g0)
            call output('gt0',gtt)
            call cal_gt_hfc(gtt,g)
            call output('gt2',gtt)
         end if
         call sc2nd(sigma,0.0001_wp,1)
         g = g0/(1.0-g0*sigma)
         if(is_use_new)then
            call cal_gt(gtt,g)
            call output('gt20',gtt)
         else
            call cal_gt_hfc(gtt,g)
            call output('gt20',gtt)
         end if
      end if

      if(sigma0_mode.eq.0)then
         sigma = 0.0
         g = g0/(1.0-g0*sigma)
      else if(sigma0_mode.eq.1)then  ! 2nd-order 
         if(rank.eq.0) call sc2nd(sigma,0.0001_wp,1)
      else if(sigma0_mode.eq.2)then   ! self-consitent 2nd-order 
         if(rank.eq.0) write(*,*)'   self-consisten 2nd-order',npt
         if(rank.eq.0) call sc2nd(sigma,0.0001_wp,npt)
      else if(sigma0_mode.eq.3)then ! using sigma_flex
         if(rank.eq.0) then
            open(unit=20,file='sigma_flex.dat',status='old')
            do n=0,nt-1
               read(20,*)sigma(n)
	    end do
	    close(20)
         end if
      else if(sigma0_mode.eq.4 .and. nc.eq.1)then !  use the exact result
	 do i=0,nf-1
	    if(is_use_new)then
	       g(i) = 0.5*(dt/(exp((w(i)-U/2.0)*dt)-1.0) 
     &                     + dt/(exp((w(i)+U/2.0)*dt)-1.0))
	    else
	       g(i) = 1.0/(w(i)-U**2/4.0/w(i))
	    end if
	 end do
      end if
     
      if(sigma0_mode.le.3)then
         call wp2mpitype(kind(sigma),dtype,ztype)
         call mpi_bcast(sigma,nt,ztype,0,mpi_comm_world,ierr)      
         if(ierr.ne.mpi_success)print *,rank, ' error in mpi_bcast'
         g = g0/(1.0-g0*sigma)
      end if




c
c          initialization of vertex functions
c

      g_d =  U/nc
      g_m = -U/nc
      g_s = 2.0*U/nc
      g_t = 0

      f_d = g_d
      f_m = g_m
      f_s = g_s
      f_t = g_t




c======================================================================
c     input the damping factors
c======================================================================
c      open (unit=1,file='indat',form='formatted',status='old')
c      read (1,*)alpha0, dfac_s, dfac_g
      if(U/T.le.2.5)then
         alpha0 = 1.0_wp
         dfac_s = 1.0_wp
         dfac_g = 1.0_wp
      else
         alpha0 = 80.0_wp
         dfac_s = 0.1_wp
         dfac_g = 0.91_wp
         !  use a very large sigma to damp the vertices in the first iteration
         sigma = alpha0*sigma
         do j=0,nc-1
	   do i=0,nf-1
              sigma(i+j*nf) = -ii*alpha0
	      if(i.ge.nf/2) sigma(i+j*nf) = ii*alpha0
	   end do
         end do
         g = g0/(1.0-g0*sigma)
      end if
      if(rank.eq.0)write(*,*)alpha0, dfac_s, dfac_g




      ! start the parquet loop
      iter = 1
      converged = .false.
      sigma_old = sigma
      do while((.not.converged .and. iter.le.iter_max))
         if(rank.eq.0)then
            print *, ' '
            print *, ' '
            write(*,'(A16,I3,A12)'),'   iteration',iter, 'starts'
            write(*,*)'     starting..........................'
         end if


         if(.not.is2nd)then
      ! form the non-perturbed susceptibilities
            chi0ph = 0.0
            chi0pp = 0.0
            do m1=0,nb-1
               m = m1+rank*nb
               do n1=0,nt-1
                  chi0ph(n1,m1) = T*g(n1)*g(add_fb(n1,m))
                  chi0pp(n1,m1) = -0.5*T*g(inv(n1))*g(add_fb(n1,m))
               end do
            end do

            if(.true.)then
      ! get the new reducible vertices by solving the Bethe-Salpeter equations
               call profstart('solve_BS_1')
               call solve_bs(f_s,g_s,chi0pp)
               call solve_bs(f_t,g_t,chi0pp)
               if(is_bs_pp .and. (iter.ne.1))then
                  call rotator4(f_s,mat1,temat)
                  call rotator4(f_t,mat2,temat)
                  f_d =  0.5*mat1 + 1.5*mat2
                  f_m = -0.5*mat1 + 0.5*mat2
               else
                  call solve_bs(f_d,g_d,chi0ph)
                  call solve_bs(f_m,g_m,chi0ph)
               end if
               call profend('solve_BS_1')
            end if

         end if


         if(.not.isflex .and. .not.is2nd)then

	!  start the inner loop
            iter_inner = 1
            converged_inner = .false.
            gm0 = g_m(0,0,0)
            gm0_old = gm0
            do while((.not.converged_inner).and.
     &               iter_inner.le.iter_max_inner)

               if(rank.eq.0)write(*,*)'            ',iter_inner
       
            call profstart('solve_parquet')
            if(use_min_memory)then
c      phi_d => mat1
c      psi_s => mat1
c      phi_m => mat2
c      psi_t => mat2
      ! calculate the kernels (vertex ladders)
               mat1 = 0.0
               mat2 = 0.0
               do m1=0,nb-1
                  m = rank*nb+m1
                  do n3=0,nt-1
                     do n2=0,nt-1
                        do n1=0,nt-1
                           mat1(n1,n2,m1) = mat1(n1,n2,m1) 
     &                                     + f_s(n1,n3,m1)
     &                                       *chi0pp(n3,m1)
     &                                       *g_s(n3,n2,m1)
                           mat2(n1,n2,m1) = mat2(n1,n2,m1) 
     &                                     + f_t(n1,n3,m1)
     &                                       *chi0pp(n3,m1)
     &                                       *g_t(n3,n2,m1)
                        end do
                     end do
                  end do
               end do

      ! rotate the kernels
               call mpi_barrier(mpi_comm_world,ierr)
               call rotator4(mat1,mat3,temat)  ! psi_s1 = mat3
               call rotator4(mat2,mat4,temat)  ! psi_t1 = mat4



               mat1 = 0.0
               mat2 = 0.0
               do m1=0,nb-1
                  m = rank*nb+m1
                  do n3=0,nt-1
                     do n2=0,nt-1
                        do n1=0,nt-1
                           mat1(n1,n2,m1) = mat1(n1,n2,m1) 
     &                                     + f_d(n1,n3,m1)
     &                                       *chi0ph(n3,m1)
     &                                       *g_d(n3,n2,m1)
                           mat2(n1,n2,m1) = mat2(n1,n2,m1) 
     &                                     + f_m(n1,n3,m1)
     &                                       *chi0ph(n3,m1)
     &                                       *g_m(n3,n2,m1)
                        end do
                     end do
                  end do
               end do


               call mpi_barrier(mpi_comm_world,ierr)
               g_d = (1.0-dfac_g)*g_d + dfac_g*(
     &                 U/Nc  + 0.5*mat3 + 1.5*mat4)
               g_m = (1.0-dfac_g)*g_m + dfac_g*(
     &                -U/Nc  - 0.5*mat3 + 0.5*mat4)


               call rotator1(mat1,mat3,temat)  ! phi_d1 = mat3
               call rotator1(mat2,mat4,temat)  ! phi_m1 = mat4

               g_d = g_d + dfac_g*(
     &                       - 0.5*mat3 - 1.5*mat4)
               g_m = g_m + dfac_g*(
     &                       - 0.5*mat3 + 0.5*mat4)



               call rotator2(mat1,mat3,temat)  ! phi_d2 = mat3
               call rotator2(mat2,mat4,temat)  ! phi_m2 = mat4
               g_s = (1.0-dfac_g)*g_s + dfac_g*(
     &              2.0*U/Nc + 0.5*mat3 - 1.5*mat4)
               g_t = (1.0-dfac_g)*g_t + dfac_g*(
     &                 0     + 0.5*mat3 + 0.5*mat4)


               call rotator3(mat1,mat3,temat)  ! phi_d3 = mat3
               call rotator3(mat2,mat4,temat)  ! phi_m3 = mat4



      ! Solve the parquet equations
               g_s = g_s + dfac_g*(
     &                       + 0.5*mat3 - 1.5*mat4)
               g_t = g_t + dfac_g*(
     &                       - 0.5*mat3 - 0.5*mat4)
            else
      ! calculate the kernels (vertex ladders)
               call profstart('cal_ladders')
               phi_d = 0.0
               phi_m = 0.0
               psi_s = 0.0
               psi_t = 0.0
	       if(.true.)then


             do m1=0,nb-1
                do n2=0,nt-1
                   fchi0_d(:,n2) = f_d(:,n2,m1)*chi0ph(n2,m1)
                   fchi0_m(:,n2) = f_m(:,n2,m1)*chi0ph(n2,m1)
                   fchi0_s(:,n2) = f_s(:,n2,m1)*chi0pp(n2,m1)
                   fchi0_t(:,n2) = f_t(:,n2,m1)*chi0pp(n2,m1)
                end do

                if(wp.eq.dp)then
                call zGEMM('N','N',nt,nt,nt,onec,fchi0_d,nt,
     &                     g_d(0,0,m1),nt,zeroc,phi_d(0,0,m1),nt)
                call zGEMM('N','N',nt,nt,nt,onec,fchi0_m,nt,
     &                     g_m(0,0,m1),nt,zeroc,phi_m(0,0,m1),nt)
                call zGEMM('N','N',nt,nt,nt,onec,fchi0_s,nt,
     &                     g_s(0,0,m1),nt,zeroc,psi_s(0,0,m1),nt)
                call zGEMM('N','N',nt,nt,nt,onec,fchi0_t,nt,
     &                     g_t(0,0,m1),nt,zeroc,psi_t(0,0,m1),nt)
                else
                call cGEMM('N','N',nt,nt,nt,onec,fchi0_d,nt,
     &                     g_d(0,0,m1),nt,zeroc,phi_d(0,0,m1),nt)
                call cGEMM('N','N',nt,nt,nt,onec,fchi0_m,nt,
     &                     g_m(0,0,m1),nt,zeroc,phi_m(0,0,m1),nt)
                call cGEMM('N','N',nt,nt,nt,onec,fchi0_s,nt,
     &                     g_s(0,0,m1),nt,zeroc,psi_s(0,0,m1),nt)
                call cGEMM('N','N',nt,nt,nt,onec,fchi0_t,nt,
     &                     g_t(0,0,m1),nt,zeroc,psi_t(0,0,m1),nt)
                end if
             end do         

	       else
               do m1=0,nb-1
                  m = rank*nb+m1
                  do n3=0,nt-1
                     do n2=0,nt-1
                        do n1=0,nt-1
                           phi_d(n1,n2,m1) = phi_d(n1,n2,m1) 
     &                                     + f_d(n1,n3,m1)
     &                                       *chi0ph(n3,m1)
     &                                       *g_d(n3,n2,m1)
                           phi_m(n1,n2,m1) = phi_m(n1,n2,m1) 
     &                                     + f_m(n1,n3,m1)
     &                                       *chi0ph(n3,m1)
     &                                       *g_m(n3,n2,m1)
                           psi_s(n1,n2,m1) = psi_s(n1,n2,m1) 
     &                                     + f_s(n1,n3,m1)
     &                                       *chi0pp(n3,m1)
     &                                       *g_s(n3,n2,m1)
                           psi_t(n1,n2,m1) = psi_t(n1,n2,m1) 
     &                                     + f_t(n1,n3,m1)
     &                                       *chi0pp(n3,m1)
     &                                       *g_t(n3,n2,m1)
                        end do
                     end do
                  end do
               end do
 	       end if
               call profend('cal_ladders')


      ! rotate the kernels
               call mpi_barrier(mpi_comm_world,ierr)
               call profstart('rotate_ladders_1')
               call rotator1(phi_d,mat1,temat)
               call rotator1(phi_m,mat2,temat)
               call rotator4(psi_s,mat3,temat)
               call rotator4(psi_t,mat4,temat)
               call profend('rotate_ladders_1')



      ! Solve the parquet equations
               call profstart('solve_pe_1')
               g_d = (1.0-dfac_g)*g_d + dfac_g*(
     &                 U/Nc  - 0.5*mat1 - 1.5*mat2
     &                       + 0.5*mat3 + 1.5*mat4)
               g_m = (1.0-dfac_g)*g_m + dfac_g*(
     &                -U/Nc  - 0.5*mat1 + 0.5*mat2
     &                       - 0.5*mat3 + 0.5*mat4)
               call profend('solve_pe_1')

               call profstart('rotate_ladders_2')
               call rotator2(phi_d,mat1,temat)
               call rotator2(phi_m,mat2,temat)
               call rotator3(phi_d,mat3,temat)
               call rotator3(phi_m,mat4,temat)
               call profend('rotate_ladders_2')

               call profstart('solve_pe_2')
               g_s = (1.0-dfac_g)*g_s + dfac_g*(
     &              2.0*U/Nc + 0.5*mat1 - 1.5*mat2
     &                       + 0.5*mat3 - 1.5*mat4)
               g_t = (1.0-dfac_g)*g_t + dfac_g*(
     &                 0     + 0.5*mat1 + 0.5*mat2
     &                       - 0.5*mat3 - 0.5*mat4)
               call profend('solve_pe_2')

            end if
               call profend('solve_parquet')

               if(rank.eq.0)gm0 = g_m(0,0,0)

      ! check whether the g_m(0,0,0) converges
               call check_converged2(converged_inner)

               if(rank.eq.0)gm0_old = gm0

               iter_inner = iter_inner + 1
               
            end do              !  do while()



      ! get the new reducible vertices by solving the Bethe-Salpeter equations
            call profstart('solve_BS_2')
            call solve_bs(f_s,g_s,chi0pp)
            call solve_bs(f_t,g_t,chi0pp)
            if(is_bs_pp)then
               call rotator4(f_s,mat1,temat)
               call rotator4(f_t,mat2,temat)
               f_d =  0.5*mat1 + 1.5*mat2
               f_m = -0.5*mat1 + 0.5*mat2
            else
               call solve_bs(f_d,g_d,chi0ph)
               call solve_bs(f_m,g_m,chi0ph)
            end if
            call profend('solve_BS_2')


      ! test the leading values of chi*Gamma	
            if(isLEV) call cal_LEV(g_d,g_m,g_s,g_t,led)

      ! test crossing symmetries	
            if(isCS) call check_crossing(f_d,f_m,f_s,f_t,val) 

         end if                 ! if(.not.isflex)


         if(is2nd)then
            f_d = g_d
            f_m = g_m
            f_s = g_s
            f_t = g_t
         end if
         


      ! update the self-engergy by using the reducible vertices
         call cal_sigma(sigma,g,f_d,f_m,f_s,f_t,sig_mode)
        

         !  damp the self energy
	sigma = dfac_s*sigma + (1.0-dfac_s)*sigma_old

      ! check whether the self-energy converges
         call check_converged(converged)

         sigma_old = sigma


      ! calculate green function by using Dyson's equation   
      ! might have some problem here since mu is not used   
         g = g0/(1.0-g0*sigma)

         if(rank.eq.0)then
            write(*,'(A16,I3,A12)'),'   iteration',iter, 'ends'
            write(*,*)'     end======================'
            write(*,*)' '
         end if


         iter = iter + 1

      end do ! do while()



      ! calculate the leading values of chi*Gamma	
      call cal_LEV(g_d,g_m,g_s,g_t,led)

      !  write out the one-particle quantities
      if(rank.eq.0)then
         open(unit=40,file="sgg.dat",status="unknown")
	 write(40,*)g0, g, sigma
	 close(40)

         open(unit=41,file="sigma.dat",status="unknown")
         open(unit=42,file="g0.dat",status="unknown")
         open(unit=43,file="g.dat",status="unknown")
	 do n=0,nt-1
	    write(41,112)n,sigma(n)
	    write(42,112)n,g0(n)
	    write(43,112)n,g(n)
         end do
	 close(41)
	 close(42)
	 close(43)
      end if



      call cal_nn                            ! double occupancy & local moment
      if(rank.eq.0) call cal_chi('chi.dat')  ! uniform susceptibility
      ! get the green's function on the Matsubara time
      if(rank.eq.0)then
         if(is_use_new)then
            call cal_gt(gtt,g)
            call output('gt',gtt)
         else
            call cal_gt_hfc(gtt,g)
            call output('gt',gtt)
         end if
      end if

      ! write out F, G to file
c      call write_fg





      if(rank.eq.0)then
         write(*,*)' '
         write(*,*)' '
         write(*,*)'         END of parquet calculations.'
         write(*,*)' '
         write(*,*)' '
      end if


      call profend('pa')
      if (rank.eq.0) then
         call profstat()
      endif

      call mpi_finalize(ierr)

 103     format((3X,f4.1),2(3X,f9.4))
 104     format(4(3X,f9.4))
 105     format((8X,f4.1),4(3X,f9.4))
 109     format((3X,f4.1),8(3X,f5.2))
 112     format((3X,I3),2(3X,f14.9))
 114     format((3X,I3),4(3X,f9.4))
 116     format((3X,I3),6(3X,f9.4))
 122     format(2(3X,I3),2(3X,f9.4))
 132     format(3(3X,I3),2(3X,f9.4))
 134     format(3(3X,I3),4(3X,f9.4))
 138     format(3(3X,I3),8(3X,f9.4))



      contains

      subroutine cal_nn
      complex(kind=wp) :: D, D_hf, temp
      !  calculate the double occupancy & the local moment
      if(rank.eq.0)then
	 D = 0.0
	 D_hf = 0.0
         do kx1=0,Nkx-1
            do ky1=0,Nky-1
               k = kx1+Nkx*ky1
	       temp = 0.0
               do n=0,nf-1
                  i = n+k*nf
	          temp = temp + g(i)*sigma(i)- 0.5*U/(w(n)-epsk(kx1,ky1))
               end do
	       D = D + T*temp/U + 0.5/(exp(epsk(kx1,ky1)/T)+1.0)
	       D_hf = D_hf + 0.5/(exp(epsk(kx1,ky1)/T)+1.0)
            end do
         end do
         D = D/Nc
         D_hf = D_hf/Nc
	 write(*,*)'   D ', D
         open(unit=21,file="D_pa.dat",status="unknown",ACCESS='append')
	 write(21,*)(T/U), real(D), real(D_hf)
	 close(21)
         open(unit=22,file="mu.dat",status="unknown",ACCESS='append')
	 write(*,*)(0.25/T), 1.0-2.0*real(D), 1.0-2.0*real(D_hf)
	 write(22,*)(0.25/T), 1.0-2.0*real(D), 0.25/T/nf
	 close(22)
      end if
      end subroutine cal_nn

      subroutine cal_chi(filename)
      !  calculate the uniform susceptibility ( structure factor (0,0) * T)
      character(len=*), intent(in) :: filename
      complex(kind=wp) :: temp, temp2

      write(*,*)' '
      write(*,*)' '
      temp = 0.0
      do ky=0,Nky-1
         do kx=0,Nkx-1
            k = kx+Nkx*ky
            do i=0,nf-1
               n = i+nf*k
               temp = temp + T**2 * (g(n)**2-g0(n)**2)
            end do
            temp = temp - exp(epsk(kx,ky)/T)/(exp(epsk(kx,ky)/T)+1.0)**2
         end do
      end do

      temp = -2.0*temp /Nc
      if(rank.eq.0)write(*,*)'  chi0: ', real(temp), imag(temp)


      temp2 = 0.0
      do n=0,nt-1 
         do n1=0,nt-1
            temp2 = temp2 + g(n)*g(n)*f_m(n,n1,0)*g(n1)*g(n1)
         end do
      end do
      temp2 = -2.0*T**3 * temp2 / Nc   !  note that f_m inclues 1/Nc
      write(*,*)'  chi_v: ', real(temp2), imag(temp2)

      
      write(*,*)'  chi: ', real(temp+temp2)/(T/0.25)
      open(unit=27,file=filename,status="unknown",ACCESS='append')
      write(27,104)(0.25/T), real(temp+temp2)/(T/0.25)
     &              ,real(temp)/(T/0.25), real(temp2)/(T/0.25) 
      close(27)

 104     format(4(3X,f14.7))
      end subroutine cal_chi





      subroutine write_fg
      integer :: m1,m,n1,n2
      integer :: OUT,etype,filetype,info=0, inc
      integer :: OUT1, OUT2
      integer (KIND=MPI_OFFSET_KIND) :: disp  
      character*40 :: fn, fmt
    
      call profstart('write_fg')

      call MPI_FILE_OPEN(MPI_COMM_WORLD, 'fg.dat', 
     &        MPI_MODE_CREATE+MPI_MODE_WRONLY, MPI_INFO_NULL, OUT, ierr)

      call MPI_FILE_OPEN(MPI_COMM_WORLD, 'fm.dat',   !  f_m
     &       MPI_MODE_CREATE+MPI_MODE_WRONLY, MPI_INFO_NULL, OUT1, ierr)

      call MPI_FILE_OPEN(MPI_COMM_WORLD, 'fst.dat',  !  f_s - f_t
     &       MPI_MODE_CREATE+MPI_MODE_WRONLY, MPI_INFO_NULL, OUT2, ierr)

      inc = nt**2
      inc = inc * nt * wp * 2

      disp=nt**2 * nb
      disp=disp * rank * wp * 2

      call write_fg_single(f_d, disp, OUT)

      call write_fg_single(f_m, disp, OUT1)
      temat = f_s-f_t
      call write_fg_single(temat, disp, OUT2)

      disp=disp + inc
      call write_fg_single(f_m, disp, OUT)
      disp=disp + inc
      call write_fg_single(f_s, disp, OUT)
      disp=disp + inc
      call write_fg_single(f_t, disp, OUT)
      disp=disp + inc
      call write_fg_single(g_d, disp, OUT)
      disp=disp + inc
      call write_fg_single(g_m, disp, OUT)
      disp=disp + inc
      call write_fg_single(g_s, disp, OUT)
      disp=disp + inc
      call write_fg_single(g_t, disp, OUT)
	 
      call MPI_FILE_CLOSE(OUT,ierr)
      call MPI_FILE_CLOSE(OUT1,ierr)
      call MPI_FILE_CLOSE(OUT2,ierr)
      call profend('write_fg')
      end subroutine write_fg

      subroutine write_fg2
      integer :: m1,m,n1,n2
      integer :: OUT, inc
      integer :: OUT1, OUT2
      integer (KIND=MPI_OFFSET_KIND) :: disp  
      character*40 :: fn, fmt
      integer :: etype,filetype,info=0
      integer :: status(mpi_status_size)
    
      call profstart('write_fg')

      if(rank.eq.0) write(*,*)'   01   ok '
      call MPI_FILE_OPEN(MPI_COMM_WORLD, 'fg.dat', 
     &        MPI_MODE_CREATE+MPI_MODE_WRONLY, MPI_INFO_NULL, OUT, ierr)

      if(rank.eq.0) write(*,*)'   02   ok '

      call MPI_FILE_OPEN(MPI_COMM_WORLD, 'fm.dat',   !  f_m
     &       MPI_MODE_CREATE+MPI_MODE_WRONLY, MPI_INFO_NULL, OUT1, ierr)

      if(rank.eq.0) write(*,*)'   03   ok '
      call MPI_FILE_OPEN(MPI_COMM_WORLD, 'fst.dat',  !  f_s - f_t
     &       MPI_MODE_CREATE+MPI_MODE_WRONLY, MPI_INFO_NULL, OUT2, ierr)

      if(rank.eq.0) write(*,*)'   04   ok '
      inc = nt**2
      inc = inc * nt * wp * 2

      disp=nt**2 * nb
      disp=disp * rank * wp * 2

      call wp2mpitype(kind(f_d),dtype,ztype)
      etype=ztype
      filetype=ztype
      
      call MPI_FILE_SET_VIEW(OUT,disp,etype,filetype, 
     &       'NATIVE', info,ierr)
      
      call MPI_FILE_SET_VIEW(OUT1,disp,etype,filetype, 
     &       'NATIVE', info,ierr)
      
      call MPI_FILE_SET_VIEW(OUT2,disp,etype,filetype, 
     &       'NATIVE', info,ierr)

c      call write_fg_single(f_d, disp, OUT)
      call MPI_FILE_WRITE(OUT,f_d,nt**2*nb,ztype,status,ierr)

      if(rank.eq.0) write(*,*)'   5   ok '
c      call write_fg_single(f_m, disp, OUT1)
      call MPI_FILE_WRITE(OUT1,f_m,nt**2*nb,ztype,status,ierr)
      if(rank.eq.0) write(*,*)'   6   ok '
      temat = f_s-f_t
      if(rank.eq.0) write(*,*)'   7   ok '
c      call write_fg_single(temat, disp, OUT2)
      call MPI_FILE_WRITE(OUT2,temat,nt**2*nb,ztype,status,ierr)

      disp=disp + inc
c      call write_fg_single(f_m, disp, OUT)
      call MPI_FILE_WRITE(OUT,f_m,nt**2*nb,ztype,status,ierr)
      disp=disp + inc
c      call write_fg_single(f_s, disp, OUT)
      call MPI_FILE_WRITE(OUT,f_m,nt**2*nb,ztype,status,ierr)
      disp=disp + inc
c      call write_fg_single(f_t, disp, OUT)
      call MPI_FILE_WRITE(OUT,f_m,nt**2*nb,ztype,status,ierr)
      disp=disp + inc
c      call write_fg_single(g_d, disp, OUT)
      call MPI_FILE_WRITE(OUT,f_m,nt**2*nb,ztype,status,ierr)
      disp=disp + inc
c      call write_fg_single(g_m, disp, OUT)
      call MPI_FILE_WRITE(OUT,f_m,nt**2*nb,ztype,status,ierr)
      disp=disp + inc
c      call write_fg_single(g_s, disp, OUT)
      call MPI_FILE_WRITE(OUT,f_m,nt**2*nb,ztype,status,ierr)
      disp=disp + inc
c      call write_fg_single(g_t, disp, OUT)
      call MPI_FILE_WRITE(OUT,f_m,nt**2*nb,ztype,status,ierr)
	 


c      call MPI_FILE_WRITE(OUT,f,nt**2*nb,ztype,status,ierr)



      call MPI_FILE_CLOSE(OUT,ierr)
      call MPI_FILE_CLOSE(OUT1,ierr)
      call MPI_FILE_CLOSE(OUT2,ierr)
      call profend('write_fg')
      end subroutine write_fg2



      subroutine write_fg_single(f, disp, OUT)
      complex(kind=wp), intent(in) :: f(0:nt-1,0:nt-1,0:nb-1)
      integer(KIND=MPI_OFFSET_KIND), intent(in) :: disp
      integer, intent(in) :: OUT
      integer :: m1,n1,n2
      integer :: etype,filetype,info=0
      integer :: status(mpi_status_size)

      call wp2mpitype(kind(f),dtype,ztype)
      etype=ztype
      filetype=ztype
      
      call MPI_FILE_SET_VIEW(OUT,disp,etype,filetype, 
     &       'NATIVE', info,ierr)

      do m1=0,-nb-1
         do n1=0,nt-1
	    do n2=0,nt-1
	       call MPI_FILE_WRITE(OUT,f(n1,n2,m1),1,ztype,status,ierr)
            end do
         end do
      end do

      call MPI_FILE_WRITE(OUT,f,nt**2*nb,ztype,status,ierr)


 132     format(3(3X,I3),2(3X,f14.8))
      end subroutine write_fg_single


      subroutine check_converged(converged)
      logical, intent(out) :: converged
      real(kind=wp) :: er, er_t
      integer :: k,n
c      call profstart('check_converged')
      if(iter.gt.1)then
         if(rank.eq.0)then
            er = 0.0
            er_t = 0.0
            do k=0,nc-1
               n = k*nf
               er = er + abs(sigma(n)-sigma_old(n))
               er_t = er_t + abs(sigma(n)+sigma_old(n))
            end do
            er = er/(0.5*er_t)
            write(*,'(a30,f12.10,a20,f12.10)')
     &           ' relative sigma diff ',er,'  required: ',er_min
            if(er.lt.er_min) converged = .true.
         end if

            call mpi_bcast(converged,1,mpi_logical,0,
     &                  mpi_comm_world,ierr)

         if(ierr.ne.mpi_success)print *, ' error in mpi_bcast'
      end if
c      call profend('check_converged')
      end subroutine check_converged


      subroutine check_converged2(converged2)
      logical, intent(out) :: converged2
      real(kind=wp) :: er, er_t
      integer :: k,n
c      call profstart('check_converged2')
c      if(iter.gt.1)then
         if(rank.eq.0)then
               er = abs(gm0-gm0_old)
               er_t = abs(gm0+gm0_old)
            er = er/(0.5*er_t)
            write(*,'(a30,f9.6,a20,f9.6)')
     &           ' relative gamma_m0 diff ',er,'  required: ',er_min2
            if(er.lt.er_min) converged2 = .true.
         end if

            call mpi_bcast(converged2,1,mpi_logical,0,
     &                  mpi_comm_world,ierr)

         if(ierr.ne.mpi_success)print *, ' error in mpi_bcast'
c      end if
c      call profend('check_converged2')
      end subroutine check_converged2


      subroutine output(filename, data)
      character(len=*), intent(in) :: filename
      complex(kind=wp), intent(in) :: data(0:nt-1)
      character*40 :: fn, fmt
      integer :: kx,ky,k,n
c      call profstart('output')
      if(nkx.lt.10 .and. nky.lt.10)then
         fmt = "('_',i1,'_',i1,'.dat')"
      end if
      do ky=0,nky-1
         do kx=0,nkx-1
            k = kx+nkx*ky
            write(fn,fmt)kx,ky
            fn = filename // fn
            open(unit=k+10,file=fn,status='unknown')
            do n=0,nf-1
               write(k+10,103)tau(n)*T,data(n+nf*k)
            end do
            if(.not.is_use_new) write(k+10,103)tau(nf)*T,-1.0-data(nf*k)
         end do
      end do
 103     format(3(3X,f9.4))
c      call profend('output')
      end subroutine output


      subroutine cal_gt(gt,g)
      complex(kind=wp), intent(out) :: gt(0:nt-1)
      complex(kind=wp), intent(in) :: g(0:nt-1)
      integer :: kx,ky,k,n,p1,p2
      gt = 0
      do ky=0,Nky-1
         do kx=0,Nkx-1
            k = kx+Nkx*ky
            do i=0,nf-1
               p1 = i+nf*k
               do n=0,nf-1
                  p2 = n+nf*k
                  gt(p1) = gt(p1) + exp(-w(n)*tau(i))*g(p2)
               end do
               gt(p1) = T*gt(p1)
            end do
         end do
      end do
      end subroutine cal_gt


      subroutine cal_gt_hfc(gt,g)
      complex(kind=wp), intent(out) :: gt(0:nt-1)
      complex(kind=wp), intent(in) :: g(0:nt-1)
      integer :: kx,ky,k,n,p1,p2
      gt = 0
      do ky=0,Nky-1
         do kx=0,Nkx-1
            k = kx+Nkx*ky
            do i=0,nf-1
               p1 = i+nf*k
               do n=0,nf-1
                  p2 = n+nf*k
                  if(real(w0(n)).gt.0.0)then
                     gt(p1) = gt(p1) + real(g(p2))*cos(w0(n)*tau(i))
     &                      + (imag(g(p2))+1.0/w0(n))*sin(w0(n)*tau(i))
                  end if
               end do
               gt(p1) = 2.0*T*gt(p1) - 0.5
            end do
         end do
      end do
      end subroutine cal_gt_hfc


      subroutine cal_LEV(g_d,g_m,g_s,g_t,eigen_value)
      complex(kind=wp), intent(in) :: g_d(0:nt-1,0:nt-1,0:nb-1)
      complex(kind=wp), intent(in) :: g_m(0:nt-1,0:nt-1,0:nb-1)
      complex(kind=wp), intent(in) :: g_s(0:nt-1,0:nt-1,0:nb-1) 
      complex(kind=wp), intent(in) :: g_t(0:nt-1,0:nt-1,0:nb-1) 
      real(kind=wp), intent(out) :: eigen_value(0:2,0:3)
      integer :: idx(0:2)
      call profstart('cal_LEV')
      if(rank.eq.0) print *,'leading eigen values for g_d*chi^0: '
      call lev(g_d,0,eigen_value(:,0),idx)
      if(rank.eq.0) print *,'leading eigen values for g_m*chi^0: '
      call lev(g_m,0,eigen_value(:,1),idx)
      if(rank.eq.0) print *,'leading eigen values for g_s*chi^0: '
      call lev(g_s,1,eigen_value(:,2),idx)
      if(rank.eq.0) print *,'leading eigen values for g_t*chi^0: '
      call lev(g_t,1,eigen_value(:,3),idx)
      if(rank.eq.0)print *,'          '
      if(idx(0).eq.1)then  ! k=(0,0)
         open(unit=34,file="lev_00.dat",status="unknown",
     &        ACCESS='append')
	 write(34,114)iter,eigen_value(0,:)
	 close(34)
      end if
      if(idx(1).eq.1)then  ! k=(pi,pi)
         open(unit=35,file="lev_pi0.dat",status="unknown",
     &        ACCESS='append')
	 write(35,114)iter,eigen_value(1,:)
	 close(35)
      end if
      if(idx(2).eq.1)then  ! k=(pi,pi)
         open(unit=36,file="lev_pipi.dat",status="unknown",
     &        ACCESS='append')
	 write(36,114)iter,eigen_value(2,:)
	 close(36)
      end if
 105     format(5(3X,f14.7))
 114     format((3X,I3),4(3X,f14.7))
      call profend('cal_LEV')
      end subroutine cal_LEV
  

  

      !   calculate the leading eigen-values of Chi*Gamma
      !   grabbed from analysis code in DCA
      subroutine lev(g_r, mode, eigen_value, idx)
      complex(kind=wp), intent(in) :: g_r(0:nt-1,0:nt-1,0:nb-1)
      integer, intent(in) :: mode
      real(kind=wp), intent(out) :: eigen_value(0:2)
      integer, intent(out) :: idx(0:2)
      integer :: info,ieig,ilead(nt)
      real(kind=wp) :: rwork(24*nt),Sval(nt)
      complex(kind=wp) :: cwork(2*nt),Umat(nt,nt),Eval(nt)	
      integer :: i,j,m,m1,ptr
      call profstart('lev')
      idx = 0  !  1: (0,0),(pi,0),(pi,pi);  0: other cases
      eigen_value = 0.0
      do m1=0,nb-1
	 m = rank*nb+m1
	 if(m.eq.p_0 .or. m.eq.p_pi0 .or. m.eq.p_pi)then
	 if(m.eq.p_0) then
	    idx(0) = 1
            ptr = 0
         end if
	 if(m.eq.p_pi0) then
	    idx(1) = 1
            ptr = 1
         end if
	 if(m.eq.p_pi) then
	    idx(2) = 1
            ptr = 2
         end if
         do i=0,nt-1
            do j=0,nt-1
               if(mode.eq.0)then
                  temat(i,j,m1) = T*g_r(i,j,m1)*g(j)*g(add_fb(j,m))  ! ph
               else
            temat(i,j,m1) = -0.5*T*g_r(i,j,m1)*g(inv(j))*g(add_fb(j,m)) ! pp
               end if
            end do
         end do	      
         if(wp.eq.dp)then
         call zgeev('N','N',nt,temat(0,0,m1),nt,Eval,Umat,nt,
     &               Umat,nt,cwork,2*nt,rwork,info)
         else
         call cgeev('N','N',nt,temat(0,0,m1),nt,Eval,Umat,nt,
     &               Umat,nt,cwork,2*nt,rwork,info)
          end if
         if(info.ne.0) then
            write(6,*) 'zgeev error: info = ',info
            stop
         endif

c         sort the eigenvalues of M
         do i=1,nt 
c            Sval(i)=(1.0-real(Eval(i)))**2
c     &             +(aimag(Eval(i)))**2
            Sval(i)=real(Eval(i))  !  pick up the one with the largest real part
         end do
         do ieig=1,nt
            ilead(ieig)=1
            do i=2,nt
               if (Sval(i).gt.Sval(ilead(ieig))) ilead(ieig)=i
            end do
            Sval(ilead(ieig))=-1.0E8
         end do

         do i=1,nt              ! now the largest Sval is the lead
              Sval(i) = real(Eval(i))
         end do


         write(*,113)m,real(Eval(ilead(1:3)))

         eigen_value(ptr) = real(Eval(ilead(1)))

      end if
      end do
      call profend('lev')
 113  format((10x,i3),3(5x,f8.4))
 115  format((10x,i3),5(5x,f8.4))
      end subroutine lev



      subroutine sc2nd(sigma,er_min,npt)
c      subroutine sc2nd(sigma,mu,n_a,er_min,npt)
      complex(kind=wp), intent(out) :: sigma(0:nt-1)
      real(kind=wp), intent(in) :: er_min
      integer, intent(in) :: npt
      integer :: iter,n
      logical :: converged
      real(kind=wp) :: er1, er2
      complex(kind=wp) :: g(0:nt-1)
      complex(kind=wp) :: sigma_old(0:nt-1)
      real(kind=wp) :: dflex = 0.3_wp,dfacflex = 0.3_wp, dotprod
      complex(kind=wp) :: dsigma(0:nc-1),delta_sigma(0:nc-1)
c      call profstart('sc2nd')
      sigma = 0.0
      g = g0

      iter = 1
      converged = .false.
      do while(.not.converged .and. iter.le.npt)
         print *, ' '
         print *, ' sc2nd: iter ', iter, 'starts'
         call cal_sigma_2nd(sigma,g,U/Nc,-U/Nc)

         !  damp the self energy
c         if(isdamp_sigma)then
         if(.true.)then
            if(iter.gt.1)then
               do k=0,nc-1
                  dsigma(k) = sigma(k*nf)-sigma_old(k*nf)
               end do
               dotprod = 0.0
               do k=0,nc-1
                  dotprod = dotprod+real(dsigma(k))*real(delta_sigma(k))
     &                             +imag(dsigma(k))*imag(delta_sigma(k))
               end do
               if(dotprod.gt.0.0)then
                  dfacflex = dfacflex+dfacflex*0.5
                  if(dfacflex.gt.0.9) dfacflex = 0.9
               else
                  dfacflex = dfacflex-dfacflex*0.5
                  if(dfacflex.lt.0.05) dfacflex = 0.05
               end if
               sigma = dfacflex*sigma + (1.0-dfacflex)*sigma_old
               print *, '   dfacflex: ', dfacflex
            end if
            delta_sigma = dsigma
         end if   
    
c         call cal_mu(sigma,mu,n_a)
c         call dyson(g,sigma,mu)
         g = g0/(1.0-g0*sigma)
         if(iter.gt.1 )then
            er1 = 0
            er2 = 0
            do n=0,nt-1
               er1 = er1 + abs(sigma(n)-sigma_old(n))
               er2 = er2 + 0.5*abs(sigma(n)+sigma_old(n))
            end do
            print *, '  er: ', er1/er2
            if(er1/er2.lt.er_min) converged = .true.
         end if
         sigma_old = sigma
         print *, ' sc2nd: iter ', iter, 'ends'
         iter = iter + 1
      end do ! do while()  
c      call profend('sc2nd')    
      end subroutine sc2nd

      subroutine cal_sigma_2nd(sigma,g,f_d,f_m)
      complex(kind=wp), intent(out) :: sigma(0:nt-1)
      complex(kind=wp), intent(in) :: g(0:nt-1)
      real(kind=wp), intent(in) :: f_d
      real(kind=wp), intent(in) :: f_m
      complex(kind=wp) :: tevet(0:nt-1)
      integer :: n,m1,m,n1
      sigma = 0.0
      tevet = 0.0
      do m=0,nt-1
         do n1=0,nt-1
            tevet(m) = tevet(m) + g(n1)*g(add_fb(n1,m))
         end do
      end do
      do n=0,nt-1
         do m=0,nt-1
            sigma(n) = sigma(n) + tevet(m)*g(dif(n,m))
         end do
         sigma(n) = sigma(n)*(f_d-f_m)
      end do
      do n=0,nt-1
         sigma(n) = - 0.5*U*T**2*sigma(n)/nc
      end do
      end subroutine cal_sigma_2nd





      subroutine cal_g0(g0)
      complex(kind=wp), intent(out) :: g0(0:nt-1)
      integer :: kx1,ky1,k,n
      real(kind=wp) :: dt
      dt = 1.0/T/nf
      do kx1=0,Nkx-1
         do ky1=0,Nky-1
            k = kx1+Nkx*ky1
            do n=0,nf-1
	       if(is_use_new)then
                  g0(n+k*nf) = 1.0/((exp(w(n)*dt)-1)/dt-epsk(kx1,ky1))
	       else
                  g0(n+k*nf) = 1.0/(w(n)-epsk(kx1,ky1))
	       end if
            end do
         end do
      end do
      end subroutine cal_g0

      subroutine cal_sigma(sigma,g,f_d,f_m,f_s,f_t,mode)
      complex(kind=wp), intent(out) :: sigma(0:nt-1)
      complex(kind=wp), intent(in) :: g(0:nt-1)
      integer, intent(in) :: mode
      complex(kind=wp), intent(in) :: f_d(0:nt-1,0:nt-1,0:nb-1)
      complex(kind=wp), intent(in) :: f_m(0:nt-1,0:nt-1,0:nb-1)
      complex(kind=wp), intent(in) :: f_s(0:nt-1,0:nt-1,0:nb-1)
      complex(kind=wp), intent(in) :: f_t(0:nt-1,0:nt-1,0:nb-1)
      complex(kind=wp) :: tevet(0:nt-1),temp
      integer :: n,m1,m,n1
      integer :: dtype, ztype
      call profstart('cal_sigma')
      sigma = 0.0 
      do n=0,nt-1
         do m1=0,nb-1
            m = rank*nb+m1
            do n1=0,nt-1
	       if(mode.eq.0)then         !        use p-h vertex only
               temp = g(n1)*g(add_fb(n1,m))*g(dif(n,m))
     &                     *(f_d(dif(n,m),n1,m1) - f_m(dif(n,m),n1,m1))
	       else if(mode.eq.1)then    !        use p-p vertex only
               temp = g(inv(n1))*g(add_fb(n1,m))*g(inv(dif(n,m)))
     &                     *(f_s(dif(n,m),n1,m1) - f_t(dif(n,m),n1,m1)) 
	       else if(mode.eq.10)then    !        FLEX
               temp = g(n1)*g(add_fb(n1,m))*g(dif(n,m))
     &                 *(f_d(dif(n,m),n1,m1) - 3.0*f_m(dif(n,m),n1,m1))
     &              + 0.5*g(inv(n1))*g(add_fb(n1,m))*g(inv(dif(n,m)))
     &                   *(f_s(dif(n,m),n1,m1) - f_t(dif(n,m),n1,m1))	
     &              - 4.0*g(n1)*g(add_fb(n1,m))*g(dif(n,m))*U/Nc
	       else                      !        use p-h & p-p vertex
               temp = 0.5*g(n1)*g(add_fb(n1,m))*g(dif(n,m))
     &                   *(f_d(dif(n,m),n1,m1) - f_m(dif(n,m),n1,m1))
     &              + 0.5*g(inv(n1))*g(add_fb(n1,m))*g(inv(dif(n,m)))
     &                   *(f_s(dif(n,m),n1,m1) - f_t(dif(n,m),n1,m1))	
	       end if  		
               sigma(n) = sigma(n) + temp
            end do
         end do
      end do
      tevet = 0.0
      call mpi_barrier(mpi_comm_world,ierr)
      
         call wp2mpitype(kind(sigma),dtype,ztype)
         call mpi_reduce(sigma,tevet,nt,ztype,mpi_sum,0,
     &                mpi_comm_world,ierr)

         call mpi_reduce(sigma,tevet,nt,ztype,mpi_sum,0,
     &                mpi_comm_world,ierr)

      if(ierr.ne.mpi_success)print *,rank, ' error in mpi_reduce'
      sigma = tevet
      call mpi_bcast(sigma,nt,ztype,0,mpi_comm_world,ierr)
      if(ierr.ne.mpi_success)print *,rank, ' error in mpi_bcast'
      do n=0,nt-1
         sigma(n) = - 0.5*U*T**2*sigma(n)/nc
      end do
      call profend('cal_sigma')
      end subroutine cal_sigma


      subroutine solve_bs(f_r,g_r,chi0)
      complex(kind=wp),intent(out) :: f_r(0:nt-1,0:nt-1,0:nb-1)
      complex(kind=wp),intent(in) :: g_r(0:nt-1,0:nt-1,0:nb-1)
      complex(kind=wp),intent(in) :: chi0(0:nt-1,0:nb-1)
      integer :: m,m1,n1,n2,n3
      real(kind=wp) :: temp
      integer :: i,j,k, info, lda,ldb,nrhs
      integer, dimension(nt) :: ipiv

      integer, parameter :: dp = kind(1.0d0)
      integer, parameter :: sp = kind(1.0)
      complex(kind=wp),dimension(0:(nt-1),0:(nt-1)) :: A,B
      complex(kind=wp), parameter :: one = 1.0_wp

!     --------------------------------------------
!     option to perform transpose copy into temporary B
!     to reuse LAPACK zgetrs
!     --------------------------------------------
      logical, parameter :: use_Btranspose = .false.



      call profstart('solve_bs')

      do m1=0,nb-1
          m = rank*nb+m1

          call profstart('solve_bs:A')
          do n2=0,nt-1
          do n3=0,nt-1
                A(n3,n2) = delta(n2-n3)-chi0(n3,m1)*g_r(n3,n2,m1)
          end do
          end do
          call profend('solve_bs:A')

          call profstart('solve_bs:getrf')
          lda = size(A,1)
          if (kind(A).eq.dp) then
             call zgetrf(nt,nt,A,lda,ipiv,info)
          else
             call cgetrf(nt,nt,A,lda,ipiv,info)
          endif
          if (info.ne.0) then
                  write(*,*) 'solve_bs: getrf info=',info
                  stop
          endif

          call profend('solve_bs:getrf')

          call profstart('solve_bs:f_r')

          if (use_Btranspose) then
!          ---------------------------------
!          B(:,:) = transpose( g_r(:,:,m1) )
!          ---------------------------------
          do j=0,nt-1
          do i=0,nt-1
            B(i,j) = g_r(j,i,m1)
          enddo
          enddo

          nrhs = nt
          ldb = size(B,1)
          call profstart('solve_bs:getrs')
          if (kind(A).eq.dp) then
             call zgetrs('T',nt,nrhs,A,lda,ipiv, B,ldb,info)
          else
             call cgetrs('T',nt,nrhs,A,lda,ipiv, B,ldb,info)
          endif

          call profend('solve_bs:getrs')

          do j=0,nt-1
          do i=0,nt-1
            f_r(i,j,m1) = B(j,i)
          enddo
          enddo

          else
!         --------------------------------------
!         Solve for f_r in place 
!         without 2 transpose copies in B
!
!         f_r(:,:,m1) =  g_r(:,:,m1) * inverse(A)
!         f_r(:,:,m1) (P*L*U) =  g_r(:,:,m1) 
!         f_r(:,:,m1) = g_r(:,:,m1)*inv(U)*inv(L)*inv(P)
!         --------------------------------------
          do j=0,nt-1
          do i=0,nt-1
           f_r(i,j,m1) = g_r(i,j,m1)
          enddo
          enddo

          call profstart('solve_bs:trsm')
          ldb = size(f_r,1)
          nrhs = nt
          if (kind(A).eq.dp) then
            call ztrsm('Right','Upper','No trans','Non unit',nt,nrhs,         &
     &                one, A,lda, f_r(0,0,m1), ldb )

            call ztrsm('Right','Lower','No trans','Unit',nt,nrhs,             &
     &                one, A,lda, f_r(0,0,m1), ldb )
          else
            call ctrsm('Right','Upper','No trans','Non unit',nt,nrhs,         &
     &                one, A,lda, f_r(0,0,m1), ldb )

            call ctrsm('Right','Lower','No trans','Unit',nt,nrhs,             &
     &                one, A,lda, f_r(0,0,m1), ldb )

          endif
          call profend('solve_bs:trsm')

          call profstart('solve_bs:swap')
c          do i=1,nt
          do i=nt,1,-1
           j = ipiv(i)
           if (j.ne.i) then
            if (wp.eq.dp) then
             call zswap(n, f_r(0,i-1,m1),1,f_r(0,j-1,m1),1)
            else
             call cswap(n, f_r(0,i-1,m1),1,f_r(0,j-1,m1),1)
            endif
           endif
          enddo

          call profend('solve_bs:swap')

          endif


          call profend('solve_bs:f_r')
      end do
      call profend('solve_bs')
      end subroutine solve_bs



  
      subroutine scatter(a,b)
      complex(kind=wp), intent(in) :: a(0:nt-1,0:nt-1,0:nb-1)
      complex(kind=wp), intent(out) :: b(0:nt-1,0:nt-1,0:nb-1)
      complex(kind=wp) :: vet(0:nt-1)
      integer :: n,i,j,colvct
      integer :: dtype, ztype

      logical, parameter :: use_mpi_vector = .false.

      call profstart('scatter')
      b = 0.0
      call wp2mpitype(kind(a), dtype, ztype )
      if (use_mpi_vector) then
        call mpi_type_vector(nb,nt,nt,ztype,colvct,ierr)
        call mpi_type_commit(colvct,ierr)
        do n=0,nb-1
         call mpi_alltoall(a(0,0,n),1,colvct,b(0,0,n),1,colvct,
     &                     mpi_comm_world,ierr)
        end do
      else

        do n=0,nb-1
          call mpi_alltoall(a(0,0,n),nt*nb,ztype,                        &
     &                      b(0,0,n),nt*nb,ztype,                        &
     &              mpi_comm_world,ierr)
        enddo
      endif

      do j=0,nt-1,nb
         do n=0,nb-1
            do i=n+1,nb-1
               vet(:) = b(:,i+j,n)
               b(:,i+j,n) = b(:,n+j,i)
               b(:,n+j,i) = vet(:)
            end do
         end do
      end do
      if (use_mpi_vector) then
        call mpi_type_free(colvct,ierr)
      endif
      call profend('scatter')
      end subroutine scatter


      subroutine rotator1(phi,phi1,temat)
      complex(kind=wp),intent(in) :: phi(0:nt-1,0:nt-1,0:nb-1)
      complex(kind=wp),intent(out) :: phi1(0:nt-1,0:nt-1,0:nb-1)
      complex(kind=wp) :: temat(0:nt-1,0:nt-1,0:nb-1)
      integer :: i,j,n
      call profstart('rotator1')
      do n=0,nb-1
         do i=0,nt-1
            do j=0,nt-1
               phi1(i,j,n) = phi(i,add_fb(i,j),n)
            end do
         end do
      end do
      call scatter(phi1,temat)
      do n=0,nb-1
         do i=0,nt-1
            do j=0,nt-1
               phi1(i,j,n) = temat(i,dif(j,i),n)
            end do
         end do
      end do
      call profend('rotator1')
      end subroutine rotator1


      subroutine rotator2(phi,phi2,temat)
      complex(kind=wp),intent(in) :: phi(0:nt-1,0:nt-1,0:nb-1)
      complex(kind=wp),intent(out) :: phi2(0:nt-1,0:nt-1,0:nb-1)
      complex(kind=wp) :: temat(0:nt-1,0:nt-1,0:nb-1)
      integer :: i,j,n,m
      call profstart('rotator2')
      do n=0,nb-1
         m=rank*nb+n
         do i=0,nt-1
            do j=0,nt-1
              phi2(i,j,n) = phi(inv(add_fb(i,m)),add_fb(i,j),n)
            end do
         end do
      end do
      call scatter(phi2,temat)
      do n=0,nb-1
         do i=0,nt-1
            do j=0,nt-1
               phi2(i,j,n) = temat(i,dif(j,i),n)
            end do
         end do
      end do
      call profend('rotator2')
      end subroutine rotator2


      subroutine rotator3(phi,phi3,temat)
      complex(kind=wp),intent(in) :: phi(0:nt-1,0:nt-1,0:nb-1)
      complex(kind=wp),intent(out) :: phi3(0:nt-1,0:nt-1,0:nb-1)
      complex(kind=wp) :: temat(0:nt-1,0:nt-1,0:nb-1)
      integer :: i,j,n,m
      call profstart('rotator3')
      do n=0,nb-1
         m=rank*nb+n
         do i=0,nt-1
            do j=0,nt-1
               phi3(i,j,n) = phi(inv(add_fb(i,m)),add_fb(i,j),n)
            end do
         end do
      end do
      call scatter(phi3,temat)
      do n=0,nb-1
         m=rank*nb+n
         do i=0,nt-1
            do j=0,nt-1
             phi3(i,j,n)=temat(inv(add_fb(i,m)),add_ff(i,add_fb(j,m)),n)
            end do
         end do
      end do
      call profend('rotator3')
      end subroutine rotator3


      subroutine rotator4(psi,psi1,temat)
      complex(kind=wp),intent(in) :: psi(0:nt-1,0:nt-1,0:nb-1)
      complex(kind=wp),intent(out) :: psi1(0:nt-1,0:nt-1,0:nb-1)
      complex(kind=wp) :: temat(0:nt-1,0:nt-1,0:nb-1)
      integer :: i,j,n,m
      call profstart('rotator4')
      do n=0,nb-1
         do i=0,nt-1
            do j=0,nt-1
               psi1(i,j,n) = psi(inv(add_fb(i,j)),inv(i),n)
            end do
         end do
      end do
      call scatter(psi1,temat)
      do n=0,nb-1
         m = rank*nb+n
         do i=0,nt-1
            do j=0,nt-1
               psi1(i,j,n) = temat(i,add_ff(i,add_fb(j,m)),n)
            end do
         end do
      end do
      call profend('rotator4')
      end subroutine rotator4

      subroutine check_crossing(f_d,f_m,f_s,f_t,val,mode)  
      integer, optional, intent(in) :: mode
      complex(kind=wp), intent(in) :: f_d(0:nt-1,0:nt-1,0:nb-1)
      complex(kind=wp), intent(in) :: f_m(0:nt-1,0:nt-1,0:nb-1)
      complex(kind=wp), intent(in) :: f_s(0:nt-1,0:nt-1,0:nb-1) 
      complex(kind=wp), intent(in) :: f_t(0:nt-1,0:nt-1,0:nb-1) 
      real(kind=wp), intent(out) :: val(1:8) 
      call profstart('check_crossing')
      if(rank.eq.0)then
         if(present(mode))then 
            write(*,*)'    crossing symmetry test for Lambda''s: '
         else 
            write(*,*)'    crossing symmetry test for F''s: '
         end if
         write(*,*)'              mode      relative error'
      end if 
      call check_CS(f_t,f_m,f_d,1,val(1))
      call check_CS(f_s,f_m,f_d,2,val(2))
      call check_CS(f_m,f_t,f_s,3,val(3))
      call check_CS(f_d,f_t,f_s,4,val(4))
      call check_CS(f_m,f_m,f_d,5,val(5))
      call check_CS(f_d,f_m,f_d,6,val(6))
      call check_CS(f_t,f_m,f_d,7,val(7))
      call check_CS(f_s,f_m,f_d,8,val(8))
      call profend('check_crossing')
      end subroutine check_crossing


      subroutine check_CS(f_r,f_r1,f_r2,mode,val)
      complex(kind=wp), intent(in) :: f_r(0:nt-1,0:nt-1,0:nb-1)
      complex(kind=wp), intent(in) :: f_r1(0:nt-1,0:nt-1,0:nb-1)
      complex(kind=wp), intent(in) :: f_r2(0:nt-1,0:nt-1,0:nb-1)
      integer, intent(in) :: mode
      real(kind=wp), intent(out) :: val
      integer :: n1,n2,m,wbar,dw,m1
      integer :: dtype, ztype
      real(kind=wp) :: er1,er2,er_total,alpha(8),beta(8)
      call profstart('check_CS')
      alpha = (/-0.5, -1.5,  0.5, 1.5,  0.5, -1.5, 0.5, -1.5/)
      beta  = (/-0.5,  0.5, -0.5, 0.5, -0.5, -0.5, 0.5,  0.5/)
      if(mode.le.4)then
         if(mode.le.2)then
            call rotator3(f_r1, mat1, temat)
            call rotator3(f_r2, mat2, temat)
         else
            call rotator4(f_r1, mat1, temat)
            call rotator4(f_r2, mat2, temat)
         end if
      else
         if(mode.le.6)then
            call rotator1(f_r1, mat1, temat)
            call rotator1(f_r2, mat2, temat)
         else
            call rotator2(f_r1, mat1, temat)
            call rotator2(f_r2, mat2, temat)
         end if
      end if
      temat = alpha(mode)*mat1 + beta(mode)*mat2
      er1 = 0.0
      er2 = 0.0
      do m=0,nb-1
         do n1=0,nt-1
            do n2=0,nt-1
               if(abs(temat(n1,n2,m)).gt.1.0e-10)then
                  er1 = er1 + abs(temat(n1,n2,m)-f_r(n1,n2,m))
                  er2 = er2 + abs(temat(n1,n2,m)+f_r(n1,n2,m))
               end if
            end do
         end do
      end do
      er1 = er1/er2
      call wp2mpitype( kind(er1), dtype, ztype )
      call mpi_reduce(er1,er_total,1,dtype,mpi_sum,0,
     &                mpi_comm_world,ierr)
      if(rank.eq.0)then
         write(*,'(A15,I3,A10,f8.4,A2)')' ', mode,' ',100*er_total,' %' 
	 val = er_total*100
      end if	
      call profend('check_CS')
      end subroutine check_CS

      subroutine readin
      integer :: lui
      character(8) :: waslin
      integer :: dtype, ztype
      if(rank.eq.0)then
         lui = 16
         if(is_single_line)then
            open(unit=lui,file='indat_single',status='old')
            read(lui,*) nkx,nky,nf,npt,iter_max,iter_max_inner,
     &                  U, T,er_min, er_min2
            close(unit=lui) 
         else
            open(unit=lui,file='indat',status='old')
            read(lui,"(1x,a8)") waslin
            read(lui,*) nkx,nky,nf
            read(lui,"(1x,a8)") waslin
            read(lui,*) npt,iter_max,iter_max_inner
            read(lui,"(1x,a8)") waslin
            read(lui,*) U, T
            read(lui,"(1x,a8)") waslin
            read(lui,*) er_min, er_min2
            close(unit=lui) 
         end if
         nc = nkx*nky
         nt = nc*nf  
      end if

      call MPI_BCAST(nkx,1,MPI_INTEGER,0,MPI_COMM_WORLD,ierr)         ! nkx
      call MPI_BCAST(nky,1,MPI_INTEGER,0,MPI_COMM_WORLD,ierr)         ! nky
      call MPI_BCAST(nc,1,MPI_INTEGER,0,MPI_COMM_WORLD,ierr)          ! nc
      call MPI_BCAST(nf,1,MPI_INTEGER,0,MPI_COMM_WORLD,ierr)          ! nf
      call MPI_BCAST(nt,1,MPI_INTEGER,0,MPI_COMM_WORLD,ierr)          ! nt
      call MPI_BCAST(npt,1,MPI_INTEGER,0,MPI_COMM_WORLD,ierr)         ! npt
      call MPI_BCAST(iter_max,1,MPI_INTEGER,0,MPI_COMM_WORLD,ierr)    ! iter_max
      call MPI_BCAST(iter_max_inner,1,MPI_INTEGER,0,MPI_COMM_WORLD,ierr)    ! iter_max_inner

      k_0 = 0           !  (kx=0, ky=0)
      k_pi0 = nkx/2     !  (kx=pi,ky=0)
      k_pi = nkx/2+nky*nkx/2  !  (kx=pi,ky=pi)
      p_0 = 0
      p_pi = k_pi*nf
      p_pi0 = k_pi0*nf

      call wp2mpitype(kind(U),dtype,ztype)
      call MPI_BCAST(U,1,dtype,0,MPI_COMM_WORLD,ierr)           ! U
      call MPI_BCAST(T,1,dtype,0,MPI_COMM_WORLD,ierr)           ! Temperature
      call MPI_BCAST(er_min,1,dtype,0,MPI_COMM_WORLD,ierr)      ! er_min
      call MPI_BCAST(er_min2,1,dtype,0,MPI_COMM_WORLD,ierr)      ! er_min2

      nb = nt/nproc
      if(nproc*nb.ne.nt)then
         if(rank.eq.0)then
            print *, ' Error!   nproc*nb != nf '
            print *, ' nproc nb nf: ',nproc, nb, nf
            print *, ' Choose suitable nproc, nb, or nf.'
         end if
         call mpi_finalize(ierr)
         stop
      end if

      if(isLEV .and. rank.eq.0)then  
         open(unit=34,file="lev_00.dat",status="unknown")
	 write(34,*)' '
	 close(34)
         open(unit=35,file="lev_pi0.dat",status="unknown")
	 write(35,*)' '
	 close(35)
         open(unit=36,file="lev_pipi.dat",status="unknown")
	 write(36,*)' '
	 close(36)
      end if
      end subroutine readin


      end program parquet

      subroutine wp2mpitype( wp, dtype,ztype)
      integer, intent(in) :: wp
      integer, intent(out):: dtype, ztype

      include 'mpif.h'
      integer, parameter :: sp = kind(1.0)
      integer, parameter :: dp = kind(1.0d0)

      if (wp.eq.sp) then
!       ----------------
!       single precision
!       ----------------
        dtype = mpi_real
        ztype = mpi_complex
      else if (wp.eq.dp) then
        dtype = mpi_double_precision
        ztype = mpi_double_complex
      else
        write(*,*) 'invalid value for wp ',wp
        stop '** wp2mpitype ** '
      endif
      return
      end

      
